#!/usr/bin/env python

"""Trait evolution subpackage.

"""

from toytree.pcm.src.traits.discrete_markov_model_sim import (
    get_markov_model,
    simulate_discrete_data,
)

from toytree.pcm.src.traits.continuous_brownian_sim import (
    simulate_continuous_brownian,
)
